/* CYPATH: st path, cycle, chordress st path, chordless cycle enumeration algorithm
            12/Feb/2002   by Takeaki Uno  e-mail:uno@nii.jp, 
    homepage:   http://research.nii.ac.jp/~uno/index.html  */

/* This program is available for only academic use, basically.
   Anyone can modify this program, but he/she has to write down 
    the change of the modification on the top of the source code.
   Neither contact nor appointment to Takeaki Uno is needed.
   If one wants to re-distribute this code, please
    refer the newest code, and show the link to homepage of 
    Takeaki Uno, to notify the news about the codes for the users.*/


#ifndef _cypath_c_
#define _cypath_c_

#include"agrhalg.c"
#include"undo.c"
#include"problem.c"

#define CYPATH_PATH 1
#define CYPATH_CYCLE 2
#define CYPATH_CHORDLESS 4

// II->itemset is used for representing the current path by nodes 
// II->add is used for representing the current path by edges 

void CYPATH_error (){
  ERROR_MES = "command explanation";
  print_err("cypath: cpCPdDqV [option] input-filename [vertex s] [vertex t] [outputfile]\n\
%%:show progress, _:no message, +:write solutions in append mode\n\
enumerate all (chordless) cycles and paths from vertex s to t\n\
c:cycle, C:chordless cycle, p:path from s to t, P:chordless path from s to t\n\
d:directed graph, D:directed graph of reverse direction\n\
[options]\n\
-l [num]:output cycles/paths with at least [num] vertices\n\
-u [num]:output cycles/paths with at most [num] vertices\n\
-r [num]:output cycles including vertex [num]\n\
-S [num]:stop after outputting [num] solutions\n\
-, [char]:give the separator of the numbers in the output\n\
-Q [filename]:replace the output numbers according to the permutation table given by [filename]\n\
# the 1st letter of input-filename cannot be '-'.\n\
# if the output file name is -, the solutions will be output to standard output.\n");
  EXIT;
}
// -w [filename]:read weights of transactions from the file\n");

/***********************************************************************/
/*  read parameters given by command line  */
/***********************************************************************/
ALIST_ID CYPATH_read_param (int argc, char *argv[], PROBLEM *PP){
  ITEMSET *II = &PP->II;
  int c=1, t=-1;
  PP->root = -1;
  if ( argc < c+2 ){ CYPATH_error (); return (-1); }

  if ( !strchr (argv[c], '_') ){ II->flag |= SHOW_MESSAGE; PP->AG.flag |= SHOW_MESSAGE; }
  if ( strchr (argv[c], '%') ) II->flag |= SHOW_PROGRESS;
  if ( strchr (argv[c], '+') ) II->flag |= ITEMSET_APPEND;
  if ( strchr (argv[c], 'P') || strchr (argv[c], 'C') ) PP->problem |= CYPATH_CHORDLESS;
  if ( strchr (argv[c], 'p') || strchr (argv[c], 'P') ) PP->problem |= CYPATH_PATH;
  else if ( strchr (argv[c], 'c') || strchr (argv[c], 'C') ) PP->problem |= CYPATH_CYCLE;
  else error ("P, p, C or c command has to be specified", EXIT);

  if ( strchr(argv[c], 'd') ) II->dir = 1;
  if ( strchr(argv[c], 'D') ) II->dir = -1;
  c++;

  while ( argv[c][0] == '-' ){
//    if ( strchr( argv[c], 'K' ) ) II->topk.end = atoi (argv[c+1]);
    switch (argv[c][1]){
      case 'l': II->lb = atoi (argv[c+1]);
      break; case 'u': II->ub = atoi(argv[c+1]);
      break; case 'r': PP->root = atoi(argv[c+1]);
//    if ( strchr( argv[c], 'U' ) ) II->frq_ub = (WEIGHT)atof(argv[c+1]);
//    if ( strchr( argv[c], 'w' ) ) PP->weight_fname = argv[c+1];
      break; case '#': II->max_solutions = atoi(argv[c+1]);
      break; case ',': II->separator = argv[c+1][0];
      break; case 'Q': PP->outperm_fname = argv[c+1];
      break; default: goto NEXT;
    }
    c += 2;
    if ( argc < c+1 ){ CYPATH_error (); return (-1); }
  }

  NEXT:;
  PP->AG.fname = argv[c];
  if ( PP->problem & CYPATH_PATH ){
    if ( argc <= c+2 ) error ("vertices s or t is not given", EXIT);
    PP->root = atoi(argv[c+1]);
    t = atoi(argv[c+2]);
    c += 2;
  }
//  if ( II->topk.end==0 ) II->frq_lb = (WEIGHT)atof(argv[c+1]);
  if ( argc>c+1 ) PP->output_fname = argv[c+1];
  
  return (t);
}


/* insert vertices from t to the current visiting vertices to the current solution (insert edges) */
void CYPATH_add_tailpath (PROBLEM *PP, int v, int t){
  AGRAPH *G = &PP->AG;
  ITEMSET *II = &PP->II;
  int e;
  for ( ; v!=t ; v=AGRAPH_INC_FROM(*G,e,II->dir) ){
    e = PP->vf[v];
    QUE_INS (II->itemset, AGRAPH_INC_FROM(*G,e,II->dir));
    QUE_INS (II->add, e);
  }
}

/* main iteration for path enumeration */
void CYPATH_st_path_iter (PROBLEM *PP, int v, int t){
  AGRAPH *G = &PP->AG;
  ITEMSET *II = &PP->II;
  int e, vv, i, pt=II->add.t;

  II->iters++;
    /* insert vertices from t to the current visiting vertices to the current solution
       (insert edges) */
  CYPATH_add_tailpath (PP, v, t);
  ITEMSET_output_itemset (II, NULL, 0);

    /* delete visit marks, then vertices on the path are deleted */
  AGRAPH_search_reinit (PP->vf, &PP->ff);
  FLOOP (i, pt, II->add.t-1){
    vv = AGRAPH_INC_TO(*G, II->add.v[i], II->dir);
    AGRAPH_node_out (G, vv, 0);
  }
  
    /* recursive calls */
  while ( pt < II->add.t ){
    e = QUEUE_ext_tail_ (&II->add);
    QUEUE_ext_tail_ ( &II->itemset);
    vv = AGRAPH_INC_TO(*G, e, II->dir);
    if ( AGRAPH_INC_FROM(*G,e,II->dir) != t )
       AGRAPH_node_recov (G, vv, 0);
    if ( II->dir ) AGRAPH_arc_out (G, e, 0);
       else AGRAPH_edge_out (G, e, 0);
    AGRAPH_BFS (G, t, PP->vf, &PP->ff, PP->dep, II->dir, II->ub-II->add.t);
    if ( PP->vf[vv] >= 0 && PP->dep[vv]+II->add.t <= II->ub)
          CYPATH_st_path_iter (PP, vv, t);
    AGRAPH_search_reinit (PP->vf, &PP->ff);  /* delete marks */
    if ( II->dir ) AGRAPH_arc_recov (G, e, 0);
       else AGRAPH_edge_recov (G, e, 0);
  }
}

/* main iteration for chordless path enumeration */
void CYPATH_chordless_st_path_iter (PROBLEM *PP, int v, int t){
  AGRAPH *G = &PP->AG;
  ITEMSET *II = &PP->II;
  int e, u, vv, vvv, i, pt=II->add.t;

  II->iters++;
    /* insert vertices from t to the current visiting vertices to the current solution
       (insert edges) */
  CYPATH_add_tailpath (PP, v, t);
  ITEMSET_output_itemset (II, NULL, 0);

    /* delete visit marks, then vertices on the path are deleted */
  AGRAPH_search_reinit (PP->vf, &PP->ff);
  FLOOP (i, pt, II->add.t-2){
    ALISTundo_ins_stop ();
    vv = AGRAPH_INC_TO(*G,II->add.v[i],II->dir);
    vvv = AGRAPH_INC_FROM(*G,II->add.v[i],II->dir);
    MALIST_DO_FORWARD (G->edge, vv, e){
      u = AGRAPH_INC(*G,e);  /* assign to u the vertex to visit */
      if ( u != vvv ) AGRAPH_node_out (G, u, 1);
    }
  }
  i=0;
  QUEUE_ext_tail_ (&II->add);
  QUEUE_ext_tail_ (&II->itemset);

    /* recursive calls */
  while ( pt < II->add.t ){
    e = QUEUE_ext_tail_ (&II->add);
    vv = QUEUE_ext_tail_ (&II->itemset);
    vv = II->itemset.v[II->itemset.t-1];
    vvv = AGRAPH_INC_FROM(*G,e,II->dir);
    if ( i!=0 ) ALISTundo (); else i=1;
    AGRAPH_node_out (G, vvv, 0);
    AGRAPH_BFS (G, t, PP->vf, &PP->ff, PP->dep, II->dir, II->ub);

    if ( PP->vf[vv] >= 0 && PP->dep[vv] + II->add.t <= II->ub )
         CYPATH_chordless_st_path_iter (PP, vv, t); 
    else AGRAPH_search_reinit (PP->vf, &PP->ff);  /* delete marks */
    AGRAPH_node_recov (G, vvv, 0);
  }
}

/* (chordless) s-t path enumeration */
void CYPATH_st_path (PROBLEM *PP, int s, int t){
  AGRAPH *G = &PP->AG;
  ITEMSET *II = &PP->II;
  PP->dep[t] = 1;
  AGRAPH_BFS(G, t, PP->vf, &PP->ff, PP->dep, II->dir, II->ub);
  II->itemset.s = II->itemset.t = 0;
  II->add.s = II->add.t = 0;
  QUE_INS (II->itemset, s);
  if ( PP->vf[s] != -1 && PP->dep[s]<=II->ub )
    if ( (PP->problem & CYPATH_CHORDLESS) == 0 ) CYPATH_st_path_iter (PP, s, t);
    else CYPATH_chordless_st_path_iter (PP, s, t);
  else AGRAPH_search_reinit (PP->vf, &PP->ff);
}

/* chordless cycle enumeration */
void CYPATH_cycle (PROBLEM *PP){
  AGRAPH *G = &PP->AG;
  ITEMSET *II = &PP->II;
  QUEUE_INT *x;
  ALIST_ID u, e, ee, t;
    /* chordless cycle enumeration, list of vertices connected by visited edges to the vertices adjacent to v */
  FLOOP (u, 0, G->node_end){
    if ( PP->root >= 0 ) u = PP->root;
    if ( II->dir ){
      MALIST_DO_FORWARD2 (G->in, u, e, ee){
        t = AGRAPH_FROM(*G, e);
        AGRAPH_arc_out (G, e, 0);
        CYPATH_st_path (PP, u, t);
        if ( PP->problem & CYPATH_CHORDLESS ){
          PP->vf[t] = G->node_num;
          QUE_INS (PP->veccand, t);
        }
      }
      if ( PP->problem & CYPATH_CHORDLESS ){
        MQUE_FLOOP (PP->veccand, x) PP->vf[*x] = -1;
        QUEUE_RMALL_ (PP->veccand);
      }
    } else {
      MALIST_DO_FORWARD2 (G->edge, u, e, ee){
        t = AGRAPH_INC(*G, e);
        AGRAPH_edge_out (G, e, 0);
        CYPATH_st_path (PP, u, t);
        if ( PP->problem & CYPATH_CHORDLESS ){
          PP->vf[t] = G->node_num;
          QUE_INS (PP->veccand, t);
        }
      }
      if ( PP->problem & CYPATH_CHORDLESS ){
        MQUE_FLOOP (PP->veccand, x) PP->vf[*x] = -1;
        QUEUE_RMALL_ (PP->veccand);
      }
    }
    if ( PP->root >= 0 ) break;
  }
}


/*  main routine: parameter read, load file, and call main iteration */
int CYPATH_main (int argc, char *argv[]){
  PROBLEM PP;
  ALIST_ID t;

  PROBLEM_init (&PP);
  t = CYPATH_read_param (argc, argv, &PP);
if ( ERROR_MES ) return (1);

  PP.AG.flag |= PP.II.dir? LOAD_ARC: LOAD_EDGE;
  PP.II.flag |= ITEMSET_ADD;
  PROBLEM_load (&PP);
if ( ERROR_MES ) return(1);
  if ( (PP.problem & CYPATH_PATH) && (PP.root<0 || PP.root>=PP.AG.node_end || t<0 || t>=PP.AG.node_end) ){
    print_err ("a vertex is out of range, s=%d,t=%d", PP.root, t); return (1);
  }
  PROBLEM_alloc (&PP, PP.AG.node_num, PP.AG.edge_num, PP.AG.edge_num, NULL, PROBLEM_VECCAND);
if ( ERROR_MES ) return (1);
  AGRAPH_search_init (&PP.AG, &PP.vf, &PP.ff, &PP.dep);

  if ( !ERROR_MES ){
    if ( PP.problem & CYPATH_PATH ) CYPATH_st_path (&PP, PP.root, t);
    if ( PP.problem & CYPATH_CYCLE ) CYPATH_cycle (&PP);
    ITEMSET_last_output (&PP.II);
  }

  PROBLEM_end (&PP);
  return (ERROR_MES?1:0);
}

/*******************************************************************************/
#ifndef _NO_MAIN_
#define _NO_MAIN_
int main (int argc, char *argv[]){
  return (CYPATH_main (argc, argv));
}
#endif
/*******************************************************************************/

#endif
